open Base
open Okdiary
open Entry

let with_db_in_mem fn =
  Oksql.with_db ":memory:" begin fun db ->
    Oksql.Migration.apply db ~migrations:Schema.migrations;
    fn db
  end

module Testable = struct
  let date = Alcotest.testable (Fmt.of_to_string Core.Date.to_string) Core.Date.equal
  let time = Alcotest.testable (Fmt.of_to_string Core.Time.to_string) Core.Time.equal
end

let test_create_or_update_creates_entry =
  Alcotest.test_case "create_or_update creates entry" `Quick begin fun () ->
    with_db_in_mem begin fun db ->
      let repo = Entry_repo.init db in
      let user_id = "testuser@testhost.mytld" in
      let date = Core.Date.of_string_iso8601_basic "20200101" ~pos:0 in
      let content = "Test Entry Content" in

      let created_low = Core.Time.(sub (now ()) Span.second) in
      let returned_entry = Entry_repo.create_or_update repo ~user_id ~date ~content in
      let created_high = Core.Time.(add (now ()) Span.second) in

      Alcotest.(check string) "user_id of returned entry"
        user_id
        returned_entry.user_id;
      Alcotest.(check Testable.date) "date of returned entry"
        date
        returned_entry.date;
      Alcotest.(check string) "content of returned entry"
        content
        returned_entry.content;
      Alcotest.(check bool) "created_at of returned entry"
        true
        Core.Time.(between ~low:created_low ~high:created_high returned_entry.created_at);
      Alcotest.(check bool) "updated_at of retuned entry"
        true
        Core.Time.(between ~low:created_low ~high:created_high returned_entry.updated_at);

      let entry_in_db = Option.value_exn (Entry_repo.get_by_user_id_and_date repo ~user_id ~date) in
      Alcotest.(check string) "user_id of entry in db "
        user_id
        entry_in_db.user_id;
      Alcotest.(check Testable.date) "date of entry in db"
        date
        entry_in_db.date;
      Alcotest.(check string) "content of entry in db"
        content
        entry_in_db.content;
      Alcotest.(check bool) "created_at of entry in db"
        true
        Core.Time.(between ~low:created_low ~high:created_high entry_in_db.created_at);
      Alcotest.(check bool) "updated_at of entry in db"
        true
        Core.Time.(between ~low:created_low ~high:created_high entry_in_db.updated_at);
    end
  end

let test_create_or_update_updates_entry =
  Alcotest.test_case "create_or_update updates entry" `Quick begin fun () ->
    with_db_in_mem begin fun db ->
      let repo = Entry_repo.init db in
      let user_id = "testuser@testhost.mytld" in
      let date = Core.Date.of_string_iso8601_basic "20200101" ~pos:0 in
      let content = "Test Entry Content" in

      let created_low = Core.Time.(sub (now ()) Span.second) in
      ignore (Entry_repo.create_or_update repo ~user_id ~date ~content);
      let created_high = Core.Time.(add (now ()) Span.second) in

      let new_content = "Test Entry New Content" in
      let updated_low = Core.Time.(sub (now ()) Span.second) in
      let updated_entry = Entry_repo.create_or_update repo ~user_id ~date ~content:new_content in
      let updated_high = Core.Time.(add (now ()) Span.second) in

      Alcotest.(check string) "user_id of updated entry"
        user_id
        updated_entry.user_id;
      Alcotest.(check Testable.date) "date of updated entry"
        date
        updated_entry.date;
      Alcotest.(check string) "content of updated entry"
        new_content
        updated_entry.content;
      Alcotest.(check bool) "created_at of updated entry"
        true
        Core.Time.(between ~low:created_low ~high:created_high updated_entry.created_at);
      Alcotest.(check bool) "updated_at of entry in db"
        true
        Core.Time.(between ~low:updated_low ~high:updated_high updated_entry.updated_at);
    end
  end

let date_from_2020_arbitrary =
  let begin_2020 = 1577836800 in
  let end_2020 = 1609459199 in
  let date_of_epoch e =
    Core.Time.Span.of_int_sec e
    |> Core.Time.of_span_since_epoch
    |> Core.Date.of_time ~zone:Core.Time.Zone.utc in
  let date_gen = QCheck.Gen.(map date_of_epoch (int_range begin_2020 end_2020)) in
  QCheck.make date_gen

let test_find_dates_with_entries =
  let user_id = "testuser@testhost.mytld" in
  let content = "Test Entry Content" in
  QCheck.Test.make QCheck.(list date_from_2020_arbitrary) ~name:"find_dates_with_entries" begin fun dates ->
    with_db_in_mem begin fun db ->
      let repo = Entry_repo.init db in
      List.iter dates ~f:(fun date -> ignore (Entry_repo.create_or_update repo ~user_id ~date ~content));
      let expected_dates = Hash_set.of_list (module Core.Date) dates in
      let actual_dates = Entry_repo.find_dates_with_entries repo ~user_id ~year:2020 in
      Hash_set.equal expected_dates actual_dates
    end
  end

let cases = [
  test_create_or_update_creates_entry;
  test_create_or_update_updates_entry;
  QCheck_alcotest.to_alcotest test_find_dates_with_entries;
]
